/*  CDVDlinuz
 *  Copyright (C) 2002-2004  CDVDlinuz Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/cdrom.h>
#include <fcntl.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <string.h>
#include <time.h>

#include "CDVD.h"
#include "interface.h"
#include "support.h"

union {
	struct cdrom_msf msf;
	unsigned char buf[CD_FRAMESIZE_RAW];
} cr;

s32 msf_to_lba(u8 m, u8 s, u8 f) {
	u32 lsn;
	lsn = f;
	lsn+=(s - 2) * 75;
	lsn+= m * 75 * 60;
	return lsn;
}

void lba_to_msf(s32 lba, u8* m, u8* s, u8* f) {
	lba += 150;
	*m = lba / (60*75);
	*s = (lba / 75) % 60;
	*f = lba % 75;
}

#define btoi(b)		((b)/16*10 + (b)%16)		/* BCD to u_char */
#define itob(i)		((i)/10*16 + (i)%10)		/* u_char to BCD */


u8 *cdbuffer;

int cdHandle = -1;
int _Thread = 0;
pthread_t thread = -1;
int readed, subqread;
s32 disctype;
s32 traystatus;

void *CdrThread(void *arg);

#define CreateThread() { \
	if (_Thread == 0) { _Thread = 1; pthread_create(&thread, NULL, CdrThread, NULL); } }

#define ExitThread() { \
	{ _Thread = 0; pthread_exit(NULL); } }

const unsigned char version  = PS2E_CDVD_VERSION;
const unsigned char revision = 0;
const unsigned char build    = 5;    // increase that with each version

static char *libraryName      = "CDVDlinuz Driver";

u32 CALLBACK PS2EgetLibType() {
	return PS2E_LT_CDVD;
}

char* CALLBACK PS2EgetLibName() {
	return libraryName;
}

u32 CALLBACK PS2EgetLibVersion2(u32 type) {
	return (version<<16)|(revision<<8)|build;
}

GtkWidget *MsgDlg;

void OnMsg_Ok() {
	gtk_widget_destroy(MsgDlg);
	gtk_main_quit();
}

void SysMessage(char *fmt, ...) {
	GtkWidget *Ok,*Txt;
	GtkWidget *Box,*Box1;
	va_list list;
	char msg[512];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);

	if (msg[strlen(msg)-1] == '\n') msg[strlen(msg)-1] = 0;

	MsgDlg = gtk_window_new (GTK_WINDOW_DIALOG);
	gtk_window_set_position(GTK_WINDOW(MsgDlg), GTK_WIN_POS_CENTER);
	gtk_window_set_title(GTK_WINDOW(MsgDlg), "CDVDlinuz Msg");
	gtk_container_set_border_width(GTK_CONTAINER(MsgDlg), 5);

	Box = gtk_vbox_new(5, 0);
	gtk_container_add(GTK_CONTAINER(MsgDlg), Box);
	gtk_widget_show(Box);

	Txt = gtk_label_new(msg);
	
	gtk_box_pack_start(GTK_BOX(Box), Txt, FALSE, FALSE, 5);
	gtk_widget_show(Txt);

	Box1 = gtk_hbutton_box_new();
	gtk_box_pack_start(GTK_BOX(Box), Box1, FALSE, FALSE, 0);
	gtk_widget_show(Box1);

	Ok = gtk_button_new_with_label("Ok");
	gtk_signal_connect (GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnMsg_Ok), NULL);
	gtk_container_add(GTK_CONTAINER(Box1), Ok);
	GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
	gtk_widget_show(Ok);

	gtk_widget_show(MsgDlg);	

	gtk_main();
}

s32 CALLBACK CDVDinit() {
	return 0;
}

s32 CALLBACK CDVDopen() {
	LoadConf();

	if (cdHandle > 0) return 0;	/* it's already open */
	cdHandle = open(conf.Dev, O_RDONLY);
	if (cdHandle != -1) { // if we can't open the cdrom we'll works as a null plugin
		ioctl(cdHandle, CDROM_LOCKDOOR, NULL);
		ioctl(cdHandle, CDROMSTART, NULL);
	} else {
		printf("CDR: Could not open %s, working as a null plugin\n", conf.Dev);
	}

	return 0;
}

void CALLBACK CDVDclose() {
	if (cdHandle < 1) return;
//	CDVDstop();
	close(cdHandle);
	cdHandle = -1;
/*	if (thread != -1) {
		pthread_cancel(thread);
		thread = -1;
	}*/
}

void CALLBACK CDVDshutdown() {
}

s32  CALLBACK CDVDreadTrack(u32 lsn, int mode) {
	if (cdHandle < 1) {
		memset(cr.buf, 0, DATA_SIZE);
		return 0;
	}

//	printf ("CDVDreadTrack %d, mode %d\n", lsn, mode);

	lba_to_msf(lsn, &cr.msf.cdmsf_min0, %cr.msf.cdmsf_sec0, %cr.msf.cdmsf_frame0);
//	cr.msf.cdmsf_min0 = (lsn / 75) / 60;
//	cr.msf.cdmsf_sec0 = (lsn / 75) % 60 + 2;
//	cr.msf.cdmsf_frame0 = lsn % 75;

//	if (ReadMode == NORMAL) {
		if (ioctl(cdHandle, CDROMREADRAW, &cr) == -1) {
			return -1;
		}
/*	} else {					// THREADED
		readed = 0xff;
		CreateThread();
	}
	subqread = 0;*/

	switch (mode) {
		case CDVD_MODE_2352: cdbuffer = cr.buf;    break;
		case CDVD_MODE_2340: cdbuffer = cr.buf+12; break;
		case CDVD_MODE_2328: cdbuffer = cr.buf+24; break;
		case CDVD_MODE_2048: cdbuffer = cr.buf+24; break;
	}

	return 0;
}

u8*  CALLBACK CDVDgetBuffer() {
/*	if (ReadMode == THREADED) {
		while (readed == 0xff) {}
		pthread_join(thread, NULL);
		if (readed == -1)
			return NULL;
	}
*/
	return cdbuffer;
}
/*
void *CdrThread(void *arg) {
	readed = ioctl(cdHandle, CDROMREADRAW, &cr);
	ExitThread();
}*/


s32  CALLBACK CDVDreadSubQ(u32 lsn, cdvdSubQ* subq) {
	struct cdrom_subchnl subq;

	if (cdHandle < 1) {
		// fake it
		u8 min, sec, frm;
		subq->ctrl		= 4;
		subq->mode		= 1;
		subq->trackNum	= itob(1);
		subq->trackIndex= itob(1);
		
		lba_to_msf(lsn, &min, &sec, &frm);
		subq->trackM	= itob(min);
		subq->trackS	= itob(sec);
		subq->trackF	= itob(frm);
		
		subq->pad		= 0;
		
		lba_to_msf(lsn + (2*75), &min, &sec, &frm);
		subq->discM		= itob(min);
		subq->discS		= itob(sec);
		subq->discF		= itob(frm);
		return 0;
	}

	subq.cdsc_format = CDROM_MSF;
	if (ioctl(cdHandle, CDROMSUBCHNL, &subq) == -1)
		return -1;
	
	subq->ctrl		= subq.cdsc_ctrl;
	subq->mode		= subq.cdsc_adr;
	subq->trackNum	= subq.cdsc_trk;
	subq->trackIndex= subq.cdsc_ind;
	
	subq->trackM	= subq.cdsc_reladdr.msf.minute;
	subq->trackS	= subq.cdsc_reladdr.msf.second;
	subq->trackF	= subq.cdsc_reladdr.msf.frame;
	
	subq->pad		= 0;
	
	subq->discM		= subq.cdsc_absaddr.msf.minute;
	subq->discS		= subq.cdsc_absaddr.msf.second;
	subq->discF		= subq.cdsc_absaddr.msf.frame;
	
	return 0;
}

s32  CDVDgetTN(cdvdTN *Buffer) {
	struct cdrom_tochdr toc;

	if (cdHandle < 1) {
		Buffer->strack = 1;
		Buffer->etrack = 1;
		return 0;
	}

	if (ioctl(cdHandle, CDROMREADTOCHDR, &toc) == -1)
		return -1;
	
	Buffer->strack = toc.cdth_trk0;	// start track
	Buffer->etrack = toc.cdth_trk1;	// end track
	
	return 0;
}

s32  CALLBACK CDVDgetTD(u8 Track, cdvdTD *Buffer) {	//track info: min,sec,frame,type
	struct cdrom_tocentry entry;

	if (cdHandle < 1) {
		return 0;
	}

	if (Track == 0)
		Track = 0xaa;			// total time
	entry.cdte_track = Track;
	entry.cdte_format = CDROM_LBA;
	
	if (ioctl(cdHandle, CDROMREADTOCENTRY, &entry) == -1)
		return -1;

	Buffer->lsn = entry.cdte_addr.lba;
	Buffer->type= entry.cdte_datamode;

	return 0;
}

s32  CALLBACK CDVDgetTOC(void* toc) {
	u8 type = CDVDgetDiskType();
	u8* tocBuff = (u8*)toc;
	
	if(	type == CDVD_TYPE_DVDV ||
		type == CDVD_TYPE_PS2DVD)
	{
		// get dvd structure format
		// scsi command 0x43
		memset(tocBuff, 0, 2048);
		// fake it
		tocBuff[ 0] = 0x04;
		tocBuff[ 1] = 0x02;
		tocBuff[ 2] = 0xF2;
		tocBuff[ 3] = 0x00;
		tocBuff[ 4] = 0x86;
		tocBuff[ 5] = 0x72;

		tocBuff[16] = 0x00;
		tocBuff[17] = 0x03;
		tocBuff[18] = 0x00;
		tocBuff[19] = 0x00;
	}
	else if(type == CDVD_TYPE_CDDA ||
			type == CDVD_TYPE_PS2CDDA ||
			type == CDVD_TYPE_PS2CD ||
			type == CDVD_TYPE_PSCDDA ||
			type == CDVD_TYPE_PSCD)
	{
		// cd toc
		// (could be replaced by 1 command that reads the full toc)
		u8 min, sec, frm;
		s32 i, err;
		cdvdTN diskInfo;
		cdvdTD trackInfo;
		memset(tocBuff, 0, 1024);
		if (CDVDgetTN(&diskInfo) == -1)	{ diskInfo.etrack = 0;diskInfo.strack = 1; }
		if (CDVDgetTD(0, &trackInfo) == -1) trackInfo.lsn = 0;
		
		tocBuff[0] = 0x41;
		tocBuff[1] = 0x00;
		
		//Number of FirstTrack
		tocBuff[2] = 0xA0;
		tocBuff[7] = itob(diskInfo.strack);
		
		//Number of LastTrack
		tocBuff[12] = 0xA1;
		tocBuff[17] = itob(diskInfo.etrack);

		//DiskLength
		lba_to_msf(trackInfo.lsn, &min, &sec, &frm);
		tocBuff[22] = 0xA2;
		tocBuff[27] = itob(min);
		tocBuff[28] = itob(sec);
		
		for (i=diskInfo.strack; i<=diskInfo.etrack; i++)
		{
			err = CDVDgetTD(i, &trackInfo);
			lba_to_msf(trackInfo.lsn, &min, &sec, &frm);
			tocBuff[i*10+30] = trackInfo.type;
			tocBuff[i*10+32] = err == -1 ? 0 : itob(i);	  //number
			tocBuff[i*10+37] = itob(min);
			tocBuff[i*10+38] = itob(sec);
			tocBuff[i*10+39] = itob(frm);
		}
	}
	else
		return -1;
	
	return 0;
}

s32  CALLBACK CDVDgetDiskType() {
	static time_t to;
	int ret;

	if (cdHandle < 1) return 0;
	if (to == time(NULL)) return disctype;

	ret = ioctl(cdHandle, CDROM_DRIVE_STATUS);
	if (ret == CDS_TRAY_OPEN || ret == CDS_NO_DISC || ret == -1) {
		disctype = CDVD_TYPE_NODISC;
	} else {
		ret = ioctl(cdHandle, CDROM_DISC_STATUS);
		if (ret == CDS_TRAY_OPEN || ret == CDS_NO_DISC) disctype = CDVD_TYPE_NODISC;

		if (ret == CDS_AUDIO) disctype = CDVD_TYPE_CDDA; else
		if (ret == CDS_MIXED) disctype = CDVD_TYPE_PS2CDDA;
		else disctype = CDVD_TYPE_PS2CD;
	}

	to = time(NULL);
	return disctype;
}

s32  CALLBACK CDVDgetTrayStatus() {
	static time_t to;
	int ret;

	if (to == time(NULL)) return traystatus;

	ret = ioctl(cdHandle, CDROM_DRIVE_STATUS);
	if (ret == CDS_TRAY_OPEN) traystatus = CDVD_TRAY_OPEN;
	else traystatus = CDVD_TRAY_CLOSE;

	to = time(NULL);
	return traystatus;
}

s32  CALLBACK CDVDctrlTrayOpen() {
	return 0;
}
s32  CALLBACK CDVDctrlTrayClose() {
	return 0;
}

GtkWidget *ConfDlg;
GtkWidget *Edit;
GtkWidget *Combo;
GList *glist;

void OnConf_Ok() {
	char *tmp;
	int i;

	tmp = gtk_entry_get_text(GTK_ENTRY(Edit));
	strcpy(conf.Dev, tmp);
	tmp = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(Combo)->entry));
	for (i = 0; i < READ_MODES; i++)
		if (!strcmp(ReadModeNames[i], tmp))
			conf.ReadMode = i;
	gtk_widget_destroy(ConfDlg);
	SaveConf();
	gtk_main_quit();
}

void OnConf_Cancel() {
	gtk_widget_destroy(ConfDlg);
	gtk_main_quit();
}

void CALLBACK CDVDconfigure() {
	int i;

	LoadConf();

	ConfDlg = create_Config();

	Combo = lookup_widget(ConfDlg, "GtkCombo_Mode");
	Edit  = lookup_widget(ConfDlg, "GtkEntry_Dev");

	gtk_entry_set_text(GTK_ENTRY(Edit), conf.Dev);

	glist = NULL;
	for (i = 0; i < READ_MODES; i++)
		glist = g_list_append(glist, ReadModeNames[i]);
	gtk_combo_set_popdown_strings(GTK_COMBO(Combo), glist);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(Combo)->entry), ReadModeNames[conf.ReadMode]);

	gtk_widget_show_all(ConfDlg);
	gtk_main();
}

GtkWidget *About;

void OnAbout_Ok() {
	gtk_widget_destroy(About);
	gtk_main_quit();
}

void CALLBACK CDVDabout() {
	About = create_About();

	gtk_widget_show_all(About);
	gtk_main();
}

s32 CALLBACK CDVDtest() {
	cdHandle = open(conf.Dev, O_RDONLY);
	if (cdHandle == -1)
		return -1;
	close(cdHandle);
	cdHandle = -1;
	return 0;
}

void LoadConf() {
	FILE *f;
	char cfg[255];

	sprintf(cfg, "%s/.PS2E/CDVDlinuz.cfg", getenv("HOME"));
	f = fopen(cfg, "r");
	if (f == NULL) {
		strcpy(conf.Dev, DEV_DEF);
		conf.ReadMode = NORMAL;
		return;
	}
	fscanf(f, "Dev = %s\n", conf.Dev);
	fscanf(f, "ReadMode = %d\n", &conf.ReadMode);
	fclose(f);
}

void SaveConf() {
	FILE *f;
	char cfg[255];

    sprintf (cfg, "%s/.PS2E", getenv("HOME"));
	mkdir(cfg, 0755);
	sprintf(cfg, "%s/.PS2E/CDVDlinuz.cfg", getenv("HOME"));
	f = fopen(cfg, "w");
	if (f == NULL)
		return;
	fprintf(f, "Dev = %s\n", conf.Dev);
	fprintf(f, "ReadMode = %d\n", conf.ReadMode);
	fclose(f);
}
